use std::ffi::OsString;
use std::fs::File;
use std::io::{Read, Write};
use std::os::{
    fd::FromRawFd,
    unix::process::{parent_id, CommandExt},
};
use std::process::{exit, Command};
use std::time::Duration;

use anyhow::{bail, Context};
use lexopt::{Arg, Parser, ValueExt as _};
use nix::sys::wait::waitpid;
use nix::unistd::{fork, pipe, setsid, ForkResult, Pid};
use tokio::time::timeout;
use zbus::fdo::NameOwnerChangedStream;
use zbus::{export::futures_util::stream::StreamExt, fdo::DBusProxy, names::BusName, Connection};

fn usage() {
    eprintln!("Usage: dbus-waiter -n NOTIFY_FD -b BUS_NAME [-t TIMEOUT] CMD [ARGS...]");
    // eprintln!("See 'man dbus-waiter' for details");
}

fn main() -> anyhow::Result<()> {
    let mut notify_fd: Option<u16> = None;
    let mut bus_name: Option<String> = None;
    let mut timeout_secs = 10u64;
    let mut cmd: Option<OsString> = None;
    let mut args: Vec<OsString> = Vec::new();

    let mut parser = Parser::from_env();
    while let Some(arg) = parser.next().context("parsing arguments")? {
        match arg {
            Arg::Short('n') | Arg::Long("notify-fd") => {
                notify_fd = Some(parser.value()?.parse()?);
            }
            Arg::Short('b') | Arg::Long("bus-name") => {
                bus_name = Some(parser.value()?.string()?);
            }
            Arg::Short('t') | Arg::Long("timeout-secs") => {
                timeout_secs = parser.value()?.parse()?;
            }
            Arg::Short('h') | Arg::Long("help") => {
                usage();
                return Ok(());
            }
            Arg::Value(value) => {
                cmd = Some(value);
                args = parser.raw_args()?.into_iter().collect();
            }
            _ => bail!(arg.unexpected()),
        }
    }
    let notify_fd = notify_fd.context("Must specify notify-fd")?;
    let bus_name = bus_name.context("Must specify bus-name")?;
    let cmd = cmd.context("Must specify cmd")?;

    let (reader, writer) = pipe().context("creating pipe")?;
    // SAFETY: The passed fd is a valid and open file descriptor.
    let reader = unsafe { File::from_raw_fd(reader) };
    let writer = unsafe { File::from_raw_fd(writer) };

    let mut cmd = Command::new(cmd);
    cmd.args(args);

    // SAFETY: The passed fd is a valid and open file descriptor.
    let notify_fd = unsafe { File::from_raw_fd(notify_fd.into()) };

    let timeout = Duration::from_secs(timeout_secs);

    // SAFETY: This is only unsafe for multithreaded programs. This context is single-thread.
    match unsafe { fork()? } {
        ForkResult::Parent { child } => parent(cmd, child, reader),
        ForkResult::Child => child(bus_name, timeout, notify_fd, writer),
    }
}

fn parent(mut cmd: Command, child: Pid, mut r: impl Read) -> anyhow::Result<()> {
    // Reap the child ourselves since the service might not deal with that.
    waitpid(child, None)?;

    // Wait for child to signal readiness.
    let mut buffer = [0; 1];
    r.read_exact(&mut buffer)?;
    if buffer != [b'\n'; 1] {
        bail!("child process did not signal readiness");
    }

    // Returns only if 'exec' calls fails:
    anyhow::bail!(cmd.exec());
}

#[tokio::main(flavor = "current_thread")]
async fn child(
    bus_name: String,
    timeout_duration: Duration,
    notify_fd: File,
    mut w: impl Write,
) -> anyhow::Result<()> {
    let parent_pid = parent_id();
    setsid()?;

    // SAFETY: This is only unsafe for multithreaded programs. This context is single-thread.
    if unsafe { fork()? }.is_parent() {
        exit(0);
    };

    let conn = Connection::session()
        .await
        .context("connecting to user d-bus")?;
    let proxy = DBusProxy::new(&conn).await.context("creating dbus proxy")?;
    let owner_changed = proxy
        .receive_name_owner_changed_with_args(&[(0, &bus_name)])
        .await?;

    // Indicate to the parent that the waiter is ready.
    w.write_all(b"\n")?;
    drop(w); // Explicit close.

    let wait_future = wait_for_bus(proxy, owner_changed, parent_pid, notify_fd);
    timeout(timeout_duration, wait_future).await?
}

async fn wait_for_bus(
    proxy: DBusProxy<'_>,
    mut owner_changed: NameOwnerChangedStream,
    parent_pid: u32,
    mut notify_fd: impl Write,
) -> anyhow::Result<()> {
    while let Some(event) = owner_changed.next().await {
        if let Some(new_owner) = event.args()?.new_owner.take() {
            let new_owner = proxy
                .get_connection_unix_process_id(BusName::Unique(new_owner))
                .await
                .context("getting pid of new bus owner")?;

            if new_owner != parent_pid {
                bail!("another process took over the specified bus");
            }

            return notify_fd
                .write_all(b"READY=1\n")
                .context("writing to readiness fd");
        }
        // Else: no owner.
    }
    bail!("listening to signals stopped; did the d-bus die?");
}
